/**
 * Compression I: RLE Compression
 * Run-length encoding (RLE) is a data compression technique which encodes data as
 * a series of runs of a repeated single character. Runs are encoded as a length, 
 * followed by the character itself. Lengths are encoded as a single ASCII digit; 
 * runs of 10 characters or more are encoded by splitting them into multiple runs.
 * 
 *  You are given the following input string:
 *      iiiiiii225RHHuu2222fddgxxxxxxxxxxxxVVVVVVVVVVVVVooooNNDDDDDDDDDGGgqqpSSSy3FFFFFF
 *  Encode it using run-length encoding with the minimum possible output length.
 * 
 *  Examples:
 *      aaaaabccc            ->  5a1b3c
 *      aAaAaA               ->  1a1A1a1A1a1A
 *      111112333            ->  511233
 *      zzzzzzzzzzzzzzzzzzz  ->  9z9z1z  (or 9z8z2z, etc.)
 */

function matchLength(str: string, start: number): number {
    const char = str.charAt(start);
    let i;
    for (i = 0; start + i < str.length; i++) {
        const element = str.charAt(i + start);
        if (element !== char) {
            return i;
        }
    }
    return i;
}

export function* RLECompression(str: string): Generator<[string, number]> {
    let i = 0;
    while (i < str.length) {
        const char = str[i];
        let v = matchLength(str, i);
        while (v > 9) {
            yield [char, 9];
            v -= 9;
            i += 9;
        }
        yield [char, v];
        i += v;
    }
    return;
}

export function RLECompressionSolve(data: string): string[] {
    return [[...RLECompression(data)].map(([ch, v]) => `${v}${ch}`).join("")]
}

//const RLETestCase = [
//["aaaaabccc",              "5a1b3c"],
//["aAaAaA",                 "1a1A1a1A1a1A"],
//["111112333",              "511233"],
//["zzzzzzzzzzzzzzzzzzz",    "9z9z1z"],
//]
//for (const [data,ans] of RLETestCase) {
//    const predict = [...RLECompression(data)].map(([ch,v])=>`${v}${ch}`).join("")
//    console.log(predict,ans);
//}

export function decompressLZSolve(data: string): string[] {
    return [decompressLZ(data)];
}

// code from
// https://github.com/mirkoconsiglio/Bitburner-scripts/blob/master/contracts/contractor.js
//
function decompressLZ(str: string): string {
    let decoded = '', type = 0, len, ref, pos, i = 0, j;
    while (i < str.length) {
        if (i > 0) type ^= 1;
        len = parseInt(str[i]);
        ref = parseInt(str[++i]);
        if (len === 0) continue;
        if (!isNaN(ref) && type === 1) {
            i++;
            for (j = 0; j < len; j++) decoded += decoded[decoded.length - ref];
        } else {
            pos = i;
            for (; i < len + pos; i++) decoded += str[i];
        }
    }
    return decoded;
}
export function compressLZSolve(str: string): string[] {
    return [compressLZ(str)];
}

export function compressLZ(str: string): string {
    // state [i][j] contains a backreference of offset i and length j
    let cur_state = Array.from(Array(10), _ => Array(10)), new_state, tmp_state, result;
    cur_state[0][1] = ''; // initial state is a literal of length 1
    for (let i = 1; i < str.length; i++) {
        new_state = Array.from(Array(10), _ => Array(10));
        const c = str[i];
        // handle literals
        for (let len = 1; len <= 9; len++) {
            const input = cur_state[0][len];
            if (input === undefined) continue;
            if (len < 9) set(new_state, 0, len + 1, input); // extend current literal
            else set(new_state, 0, 1, input + '9' + str.substring(i - 9, i) + '0'); // start new literal
            for (let offset = 1; offset <= Math.min(9, i); offset++) { // start new backreference
                if (str[i - offset] === c) set(new_state, offset, 1, input + len + str.substring(i - len, i));
            }
        }
        // handle backreferences
        for (let offset = 1; offset <= 9; offset++) {
            for (let len = 1; len <= 9; len++) {
                const input = cur_state[offset][len];
                if (input === undefined) continue;
                if (str[i - offset] === c) {
                    if (len < 9) set(new_state, offset, len + 1, input); // extend current backreference
                    else set(new_state, offset, 1, input + '9' + offset + '0'); // start new backreference
                }
                set(new_state, 0, 1, input + len + offset); // start new literal
                // end current backreference and start new backreference
                for (let new_offset = 1; new_offset <= Math.min(9, i); new_offset++) {
                    if (str[i - new_offset] === c) set(new_state, new_offset, 1, input + len + offset + '0');
                }
            }
        }
        tmp_state = new_state;
        new_state = cur_state;
        cur_state = tmp_state;
    }
    for (let len = 1; len <= 9; len++) {
        let input = cur_state[0][len];
        if (input === undefined) continue;
        input += len + str.substring(str.length - len, str.length);
        // noinspection JSUnusedAssignment
        if (result === undefined || input.length < result.length) result = input;
    }
    for (let offset = 1; offset <= 9; offset++) {
        for (let len = 1; len <= 9; len++) {
            let input = cur_state[offset][len];
            if (input === undefined) continue;
            input += len + '' + offset;
            if (result === undefined || input.length < result.length) result = input;
        }
    }
    return result ?? '';

/**
 *
 * @param {string[][]} state
 * @param {number} i
 * @param {number} j
 * @param {string} str
 */
    function set(state: string[][], i:number, j:number, str:string): void {
        if (state[i][j] === undefined || str.length < state[i][j].length) state[i][j] = str;
    }
}